/***************************************************************************\
  "inputbuffer.h" - Input Buffer Module (Phase A)
  Revision 9; 10 Nov 2002 - 21 Feb 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht

  References:
    MCD 2.1.1; 2.1.1.1; 2.1.12.1 (Input Buffer Stack)
\***************************************************************************/

#ifndef INPUTBUFFER_H
#define INPUTBUFFER_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

// External Defines //
// GLOBAL_FILENAME_EXT

#define INPUTBUFFER_EOF             0x7F
#define INPUTBUFFER_DIR_DELIM       '\\'

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdio.h>

// Local //
#include "k_string.h"
#include "error.h"
#include "global.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// External //
void ib_init ();
void ib_deinit ();
void ib_include (char *aFilename);
char ib_get ();
void ib_seek ();

// Internal //
char mPopLine ();

typedef struct Buffer_tag
{
  FILE *file; // input file
  char *filename; // filename
  char text[GLOBAL_INPUTBUFFER_LINE_LEN]; // text buffer
  int pos; // current position in mText
  char curr; // current character
  int line; // current line number in file

  struct Buffer_tag *next; // pointer to the next buffer in the stack
} Buffer;

void mBufferFree (Buffer *buffer);
Buffer *mBufferStackTop;
char *mDirectories[GLOBAL_INPUTBUFFER_MAX_DIRECTORIES]; // array of all the directories included
int miDirectories;

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// ib_init
//// Initialize the Input Buffer Module
void ib_init ()
{
  mBufferStackTop = NULL;
  mDirectories[0] = NULL;
  miDirectories = 1;
}

//// ib_deinit
//// Deinitialize the input buffer functions
void ib_deinit ()
{
  Buffer *temp;

  // free the buffer stack //
  while (mBufferStackTop)
  {
    temp = mBufferStackTop->next;
    mBufferFree (mBufferStackTop);
    mBufferStackTop = temp;
  }

  for (; miDirectories > 0; miDirectories--)
  {
    k_del (mDirectories[miDirectories]);
  }
}

//// ib_include
//// Include a new file
void ib_include (char *aFilename)
{
  Buffer *tempnew;
  char *sTemp;
  int i, iLast;

  // if there is already a file open, set the line number //
  if (mBufferStackTop)
  {
    mBufferStackTop->line = gnCurrentLine;
  }

  // reset the current line and file information //
  gnCurrentLine = 1;
  k_copy (gsCurrentFilename, aFilename);

  // Create new buffer //
  tempnew = (Buffer *) KALLOC(sizeof (Buffer));

  // Find last directory delimiter in filename //
  iLast = -1;
  i = k_posc (aFilename, INPUTBUFFER_DIR_DELIM, 0);
  while (i != -1)
  {
    iLast = i;
    i = k_posc (aFilename, INPUTBUFFER_DIR_DELIM, i + 1);
  }

  // There is a directory delimiter in the filename //
  if (iLast != -1)
  {
    // Add the included file's directory to the list //
    mDirectories[miDirectories++] = k_sub (aFilename, 0, iLast + 1);
  }

  // Assign values to new buffer //
  tempnew->filename = NULL;
  for (i = 0; i < miDirectories; i++)
  {
    if (tempnew->filename != NULL) k_del (tempnew->filename);
    if (mDirectories[i] != NULL)
    {
      tempnew->filename = k_new (strlen (aFilename) + strlen (mDirectories[i]));
      k_copy (tempnew->filename, mDirectories[i]);
      k_cat (tempnew->filename, aFilename);
    }
    else
    {
      tempnew->filename = k_new (strlen (aFilename));
      k_copy (tempnew->filename, aFilename);
    }
    tempnew->file = fopen (tempnew->filename, "r");

    if (tempnew->file) break;
    else
    {
      // Allow files to be specified without using the extension //
      sTemp = tempnew->filename;
      tempnew->filename = k_new (strlen (sTemp) + 4);
      k_copy (tempnew->filename, sTemp);
      k_cat (tempnew->filename, GLOBAL_FILENAME_EXT);
      tempnew->file = fopen (tempnew->filename, "r");
      k_del (sTemp);

      if (tempnew->file) break;
    }
  }

  if (!tempnew->file)
  {
    err (false, EP_FILE_OPEN_FAILED, aFilename);
    errm ("Also attempted with extension", NULL);
    erre (EFF_TERMINATE);
  }

  k_copy (tempnew->text, "");
  tempnew->curr = '\0';
  tempnew->pos = 0;
  
  // put the new filename in the global filename string //
  k_copy (gsCurrentFilename, tempnew->filename);

  // Put the new buffer on top of the stack //
  tempnew->next = mBufferStackTop;
  mBufferStackTop = tempnew;

  tempnew->curr = mPopLine ();
}

//// ib_get
//// Return the current character
char ib_get ()
{
  if (!mBufferStackTop) return 0;

  return mBufferStackTop->curr;
}

//// ib_seek
//// Set the current character to the next character
void ib_seek ()
{
  if (!mBufferStackTop) return;

  if (mBufferStackTop->curr == INPUTBUFFER_EOF) // currently at end of file
  {
    return;
  }
  else if (mBufferStackTop->curr == 0) // currently at end of line
  {
    mBufferStackTop->curr = mPopLine ();
  }
  else
  {
    mBufferStackTop->curr = mBufferStackTop->text[++mBufferStackTop->pos];
  }
}

//// mPopLine
//// Put the next line from the file into the buffer.
//// Then return the line's first character.
char mPopLine ()
{
  char *sTemp; // test to see if 'fgets' was successful
  Buffer *bufferTemp;

  if (!mBufferStackTop) return '\0';

  mBufferStackTop->pos = 0;

  if (feof (mBufferStackTop->file))
  {
    // This is only the end of the currently included file //
    if (mBufferStackTop->next)
    {
      bufferTemp = mBufferStackTop;
      mBufferStackTop = mBufferStackTop->next;
      mBufferFree (bufferTemp);

      // set the current line and file inforamtion //
      gnCurrentLine = mBufferStackTop->line - 1;
      k_copy (gsCurrentFilename, mBufferStackTop->filename);

      return mBufferStackTop->text[mBufferStackTop->pos];
    }
    // This is the end the entire program //
    else
    {
      return INPUTBUFFER_EOF;
    }
  }
  else
  {
    sTemp = fgets (mBufferStackTop->text, GLOBAL_INPUTBUFFER_LINE_LEN, mBufferStackTop->file);

    //BUGFIX (14 July 2002 -- 0061):
    //  If the file was not terminated by a newline, the last char was erased
    if (mBufferStackTop->text[strlen(mBufferStackTop->text) - 1] < '\n')
    {
      mBufferStackTop->text[strlen(mBufferStackTop->text) - 1] = '\0';
    }

    if (mBufferStackTop->text && sTemp) return mBufferStackTop->text[mBufferStackTop->pos];
	  else return 0;
  }
}

//// mBufferFree
//// Free the buffer
void mBufferFree (Buffer *buffer)
{
  if (buffer)
  {
    k_del (buffer->filename);
    fclose (buffer->file);
    KFREE(buffer);
  }
}

#endif
